#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 João Pedro Rodrigues
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Renames atoms sequentially (C1, C2, O1, ...) for each HETATM residue.

Relies on an element column being present (see pdb_element).

Usage:
    python pdb_uniqname.py <pdb file>

Example:
    python pdb_uniqname.py 1CTF.pdb

This program is part of the `pdb-tools` suite of utilities and should not be
distributed isolatedly. The `pdb-tools` were created to quickly manipulate PDB
files using the terminal, and can be used sequentially, with one tool streaming
data to another. They are based on old FORTRAN77 code that was taking too much
effort to maintain and compile. RIP.
"""

import collections
import os
import sys

__author__ = ["Joao Rodrigues"]
__email__ = ["j.p.g.l.m.rodrigues@gmail.com"]


def check_input(args):
    """Checks whether to read from stdin/file.
    """

    # Defaults
    fh = sys.stdin  # file handle

    if not len(args):
        # Reading from pipe
        if sys.stdin.isatty():
            sys.stderr.write(__doc__)
            sys.exit(1)

    elif len(args) == 1:
        # Input File
        if not os.path.isfile(args[0]):
            emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
            sys.stderr.write(emsg.format(args[0]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        fh = open(args[0], 'r')

    else:  # Whatever ...
        sys.stderr.write(__doc__)
        sys.exit(1)

    return fh


def run(fhandle):
    """
    Rename HETATM atoms on each residue based on their element.

    This function is a generator.

    Parameters
    ----------
    fhandle : a line-by-line iterator of the original PDB file.

    Yields
    ------
    str (line-by-line)
        The modified (or not) PDB line.
    """
    prev_res = None
    for line_idx, line in enumerate(fhandle):
        if line.startswith('HETATM'):

            element = line[76:78].strip()

            if not element:
                emsg = 'ERROR!! No element found in line {}'.format(line_idx)
                sys.stderr.write(emsg)
                sys.exit(1)

            resuid = line[17:27]

            if prev_res != resuid:
                prev_res = resuid
                element_idx = collections.defaultdict(lambda: 1)  # i.e. a counter

            spacer = ' ' if len(element) == 1 else ''
            name = (spacer + element + str(element_idx[element])).ljust(4)
            line = line[:12] + name + line[16:]

            element_idx[element] += 1

        yield line


rename_atoms = run


def main():
    # Check Input
    pdbfh = check_input(sys.argv[1:])

    # Do the job
    new_pdb = run(pdbfh)

    try:
        _buffer = []
        _buffer_size = 5000  # write N lines at a time
        for lineno, line in enumerate(new_pdb):
            if not (lineno % _buffer_size):
                sys.stdout.write(''.join(_buffer))
                _buffer = []
            _buffer.append(line)

        sys.stdout.write(''.join(_buffer))
        sys.stdout.flush()
    except IOError:
        # This is here to catch Broken Pipes
        # for example to use 'head' or 'tail' without
        # the error message showing up
        pass

    # last line of the script
    # We can close it even if it is sys.stdin
    pdbfh.close()
    sys.exit(0)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 João Pedro Rodrigues
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deletes a range of residues from a PDB file.

The range option has three components: start, end, and step. Start and end
are optional and if ommitted the range will start at the first residue or
end at the last, respectively. The step option can only be used if both start
and end are provided. Note that the start and end values of the range are
purely numerical, while the range actually refers to every N-th residue,
regardless of their sequence number.

Usage:
    python pdb_delres.py -[resid]:[resid]:[step] <pdb file>

Example:
    python pdb_delres.py -1:10 1CTF.pdb # Deletes residues 1 to 10
    python pdb_delres.py -1: 1CTF.pdb # Deletes residues 1 to END
    python pdb_delres.py -:5 1CTF.pdb # Deletes residues from START to 5.
    python pdb_delres.py -::5 1CTF.pdb # Deletes every 5th residue
    python pdb_delres.py -1:10:5 1CTF.pdb # Deletes every 5th residue from 1 to 10

This program is part of the `pdb-tools` suite of utilities and should not be
distributed isolatedly. The `pdb-tools` were created to quickly manipulate PDB
files using the terminal, and can be used sequentially, with one tool streaming
data to another. They are based on old FORTRAN77 code that was taking too much
effort to maintain and compile. RIP.
"""

import os
import sys

__author__ = "Joao Rodrigues"
__email__ = "j.p.g.l.m.rodrigues@gmail.com"


def check_input(args):
    """Checks whether to read from stdin/file and validates user input/options.
    """

    def is_integer(string):
        """Returns True if the string contains *any* integer"""
        try:
            int(string)
            return True
        except ValueError:
            return False

    # Defaults
    option = ':::'
    fh = sys.stdin  # file handle

    if not len(args):
        # Reading from pipe with default option
        if sys.stdin.isatty():
            sys.stderr.write(__doc__)
            sys.exit(1)

    elif len(args) == 1:
        # One of two options: option & Pipe OR file & default option
        if args[0].startswith('-'):
            option = args[0][1:]
            if sys.stdin.isatty():  # ensure the PDB data is streamed in
                emsg = 'ERROR!! No data to process!\n'
                sys.stderr.write(emsg)
                sys.stderr.write(__doc__)
                sys.exit(1)

        else:
            if not os.path.isfile(args[0]):
                emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
                sys.stderr.write(emsg.format(args[0]))
                sys.stderr.write(__doc__)
                sys.exit(1)

            fh = open(args[0], 'r')

    elif len(args) == 2:
        # Two options: option & File
        if not args[0].startswith('-'):
            emsg = 'ERROR! First argument is not an option: \'{}\'\n'
            sys.stderr.write(emsg.format(args[0]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        if not os.path.isfile(args[1]):
            emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
            sys.stderr.write(emsg.format(args[1]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        option = args[0][1:]
        fh = open(args[1], 'r')

    else:  # Whatever ...
        sys.stderr.write(__doc__)
        sys.exit(1)

    # Validate option
    if not (1 <= option.count(':') <= 2):
        emsg = 'ERROR!! Residue range must be in \'a:z:s\' where a and z are '
        emsg += 'optional (default to first residue and last respectively), and'
        emsg += 's is an optional step value (to return every s-th residue).\n'
        sys.stderr.write(emsg)
        sys.exit(1)

    start, end, step = None, None, 1
    slices = [num if num.strip() else None for num in option.split(':')]
    if len(slices) == 3:
        start, end, step = slices
    elif len(slices) == 2:
        start, end = slices
    elif len(slices) == 1:
        if option.startswith(':'):
            end = slices[0]
        elif option.endswith(':'):
            start = slices[0]

    if start is None:
        start = -1000  # residues cannot reach this value (max 4 char)
    elif not is_integer(start):
        emsg = 'ERROR!! Starting value must be numerical: \'{}\'\n'
        sys.stderr.write(emsg.format(start))
        sys.exit(1)
    elif not (-999 <= int(start) < 9999):
        emsg = 'ERROR!! Starting value must be between -999 and 9998: \'{}\'\n'
        sys.stderr.write(emsg.format(start))
        sys.exit(1)

    if end is None:
        end = 10000  # residues cannot reach this value (max 4 char)
    elif not is_integer(end):
        emsg = 'ERROR!! End value must be numerical: \'{}\'\n'
        sys.stderr.write(emsg.format(end))
        sys.exit(1)
    elif not (-999 <= int(end) < 9999):
        emsg = 'ERROR!! End value must be between -999 and 9998: \'{}\'\n'
        sys.stderr.write(emsg.format(end))
        sys.exit(1)

    if step is None:
        step = 1
    elif not is_integer(step):
        emsg = 'ERROR!! Step value must be numerical: \'{}\'\n'
        sys.stderr.write(emsg.format(step))
        sys.exit(1)
    elif int(step) <= 0:
        emsg = 'ERROR!! Step value must be a positive number: \'{}\'\n'
        sys.stderr.write(emsg.format(step))
        sys.exit(1)

    start, end, step = int(start), int(end), int(step)

    if start >= end:
        emsg = 'ERROR!! Start ({}) cannot be larger than end ({})\n'
        sys.stderr.write(emsg.format(start, end))
        sys.exit(1)

    resrange = set(range(start, end + 1))
    return (fh, resrange, step)


def run(fhandle, residue_range, step):
    """
    Delete residues within a certain numbering range.

    This function is a generator.

    Parameters
    ----------
    fhandle : a line-by-line iterator of the original PDB file.

    residue_range : set, list, or tuple
        The residues describing the range.

    step : int
        The step at which delete.

    Yields
    ------
    str (line-by-line)
        All lines except RECORDS within the residue range.
    """
    prev_res = None
    res_counter = -1
    records = ('ATOM', 'HETATM', 'TER', 'ANISOU')
    for line in fhandle:
        if line.startswith(records):

            res_id = line[21:26]  # include chain ID
            if res_id != prev_res:
                prev_res = res_id
                res_counter += 1

            if int(line[22:26]) in residue_range and res_counter % step == 0:
                continue

        yield line


delete_residues = run


def main():
    # Check Input
    pdbfh, resrange, step = check_input(sys.argv[1:])

    # Do the job
    new_pdb = run(pdbfh, resrange, step)

    try:
        _buffer = []
        _buffer_size = 5000  # write N lines at a time
        for lineno, line in enumerate(new_pdb):
            if not (lineno % _buffer_size):
                sys.stdout.write(''.join(_buffer))
                _buffer = []
            _buffer.append(line)

        sys.stdout.write(''.join(_buffer))
        sys.stdout.flush()
    except IOError:
        # This is here to catch Broken Pipes
        # for example to use 'head' or 'tail' without
        # the error message showing up
        pass

    # last line of the script
    # We can close it even if it is sys.stdin
    pdbfh.close()
    sys.exit(0)


if __name__ == '__main__':
    main()



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import listdir, path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Collect names of bin/*py scripts
# e.g. 'pdb_intersect=bin.pdb_intersect:main',
binfiles = listdir(path.join(here, 'pdbtools'))
bin_py = [f[:-3] + '=pdbtools.' + f[:-3] + ':main' for f in binfiles
          if f.endswith('.py')]

setup(
    name='pdb-tools-removed-atom-limit',  # Required
    version='2.4.5',  # Required
    description='A swiss army knife for PDB files, customized to avoid the original atom limit.',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='http://bonvinlab.org/pdb-tools',  # Optional
    author='Alex Morehead',  # Optional
    author_email='alex.morehead@gmail.com',  # Optional
    license='Apache Software License, version 2',
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',
        #'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Bio-Informatics',


        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='bioinformatics protein structural-biology pdb',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(),  # Required

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        'console_scripts': bin_py,
    },

    # scripts=[path.join('bin', f) for f in listdir(path.join(here, 'bin'))],

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/amorehead/pdb-tools/issues',
        'Source': 'https://github.com/amorehead/pdb-tools',
    },

    # Test Suite
    test_suite='tests.test_all',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2018 João Pedro Rodrigues
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Checks all models in a multi-model PDB file have the same composition.

Composition is defined as same atoms/residues/chains.

Usage:
    python pdb_chkensemble.py <pdb file>

Example:
    python pdb_chkensemble.py 1CTF.pdb

This program is part of the `pdb-tools` suite of utilities and should not be
distributed isolatedly. The `pdb-tools` were created to quickly manipulate PDB
files using the terminal, and can be used sequentially, with one tool streaming
data to another. They are based on old FORTRAN77 code that was taking too much
effort to maintain and compile. RIP.
"""

import itertools
import os
import sys

__author__ = "Joao Rodrigues"
__email__ = "j.p.g.l.m.rodrigues@gmail.com"


def check_input(args):
    """Checks whether to read from stdin/file and validates user input/options.
    """

    # Defaults
    fh = sys.stdin  # file handle

    if not len(args):
        # Reading from pipe with default option
        if sys.stdin.isatty():
            sys.stderr.write(__doc__)
            sys.exit(1)

    elif len(args) == 1:
        if not os.path.isfile(args[0]):
            emsg = 'ERROR!! File not found or not readable: \'{}\'\n'
            sys.stderr.write(emsg.format(args[0]))
            sys.stderr.write(__doc__)
            sys.exit(1)

        fh = open(args[0], 'r')

    else:  # Whatever ...
        emsg = 'ERROR!! Script takes 1 argument, not \'{}\'\n'
        sys.stderr.write(emsg.format(len(args)))
        sys.stderr.write(__doc__)
        sys.exit(1)

    return fh


def run(fhandle):
    """
    Check if the ensemble is valid.

    - Same atoms in each model
    - Paired MODEL/ENDMDL tags

    Parameters
    ----------
    fhandle : a line-by-line iterator of the original PDB file.

    Returns
    -------
    int
        1 if an error was found. 0 if no errors are found.
    """
    model_open = False
    model_no = None
    model_data = {}  # list of sets
    records = ('ATOM', 'HETATM', 'ANISOU', 'TER')
    for lineno, line in enumerate(fhandle):
        if line.startswith('MODEL'):
            if model_open:
                emsg = 'ERROR!! MODEL record found before ENDMDL at line \'{}\'\n'
                sys.stderr.write(emsg.format(lineno))
                return 1

            model_open = True
            model_no = int(line[10:14])
            model_data[model_no] = set()

        elif line.startswith('ENDMDL'):
            if not model_open:
                emsg = 'ERROR!! ENDMDL record found before MODEL at line \'{}\'\n'
                sys.stderr.write(emsg.format(lineno))
                return 1

            model_open = False
            model_no = None  # will fail to add lines if a new model is not open

        elif line.startswith(records):
            if not model_open:
                emsg = 'ERROR!! MODEL record missing before ATOM at line \'{}\'\n'
                sys.stderr.write(emsg.format(lineno))
                return 1

            atom_uid = line[6:27]
            model_data[model_no].add(atom_uid)
        else:
            if model_open:  # Missing last ENDMDL
                emsg = 'ERROR!! ENDMDL record missing at line \'{}\'\n'
                sys.stderr.write(emsg.format(lineno))
                return 1

    # Verify all models have the same atoms
    bad_ensemble = False
    model_no_list = list(model_data.keys())
    for model_i, model_j in itertools.combinations(model_no_list, 2):
        atoms_i = model_data[model_i]
        atoms_j = model_data[model_j]

        difference_ij = atoms_i - atoms_j
        difference_ji = atoms_j - atoms_i

        if difference_ij or difference_ji:
            bad_ensemble = True
            msg = 'Models {} and {} differ:\n'
            sys.stderr.write(msg.format(model_i, model_j))

            if len(difference_ij):
                d_ij = sorted(difference_ij)
                msg = 'Atoms in model {} only:\n'
                sys.stderr.write(msg.format(model_i))
                sys.stderr.write('\n'.join(d_ij))

            if len(difference_ji):
                d_ji = sorted(difference_ji)
                msg = 'Atoms in model {} only:\n'
                sys.stderr.write(msg.format(model_j))
                sys.stderr.write('\n'.join(d_ji))

    if not bad_ensemble:
        n_models = len(model_data)
        msg = 'Ensemble of {} models *seems* OK\n'
        sys.stdout.write(msg.format(n_models))
        return 0

    return 1


check_ensemble = run


def main():
    # Check Input
    pdbfh = check_input(sys.argv[1:])

    # Do the job
    status_code = check_ensemble(pdbfh)

    pdbfh.close()
    sys.exit(status_code)


if __name__ == '__main__':
    main()
